; ----------------------------------------------------------------------+
; FROTH - A forth language implementation on PC - Oliver SINGLA 	!
; 									!
; Module:	THREAD.ASM						!
; 									!
; Description:	Definitions for playing with threads.			!
; 									!
; List of FROTH definitions defined in this module:			!
;	MAXNBTHREADS	MaxNbThreads					!
;	NBTHREADS	NbThreads					!
;	ACTIVETHREAD	ActiveThread					!
;	NEWTHREAD	NewThread					!
;	ENDTHREAD	EndThread					!
;	KILLTHREAD	KillThread					!
;	YIELD		Yield						!
; 									!
; ----------------------------------------------------------------------+

		.386p
		jumps
		locals

code32 		segment para public use32
		assume cs:code32, ds:code32, ss:code32

include		pmode.inc				; All externs are defined in there

include		Equ.inc
include		Vars.inc
include		Dict.inc

include		Compile.inc				; FROTH compiled definitions

include		Ed.inc                     		; Constants, Structurs, Variables used by FROSTED


; ===============================================================
; = MAXNBTHREADS						=
; =								=
; = Code for the forth word:    MaxNbThreads			=
; =								=
; = Stack:  ( --- n )	                 			=
; =								=
; = Usage:							=
; =	Return the max number of threads.			=
; ===============================================================

		align	4
MAXNBTHREADS	proc

		; Integer run-time
		DOINT	D_MAXNBTHREADS
		
		; Read data
		DO_AT

		; Verify stack frame overflow, then return to caller
		OVSTACK
		
		; Return to caller
		ret

E_MAXNBTHREADS:
MAXNBTHREADS	endp


; ===============================================================
; = ResetThreads						=
; =								=
; = Usage:							=
; = 	Reset all threads other than thread #0			=
; =								=
; = I) none                         				=
; =								=
; = O) none (never return)                  			=
; =								=
; = M) never return                     			=
; ===============================================================

		align	4
ResetThreads	proc

		; -------------------------
		; Only one thread is active
		; -------------------------

		; Nb of active threads
		mov	[ActiveThreadsNb], 1

		; Current active thread
		mov	[CurrActThread], 0
		
		; 1st pass flag: only signifiant for 2nd thread and others
		mov	[Threads1stPass], FALSE

		; Only the 1st thread is active
		lea	edx, ThreadsActive
		mov	byte ptr [edx], TRUE
		mov	ecx, 1
@@Threads1:	cmp	cx, MAXNB_THREADS
		je	@@Threads2
		mov	byte ptr [edx + ecx], FALSE
		inc	ecx
		jmp	@@Threads1
@@Threads2:

		; Return to caller
		ret

E_ResetThreads:		
ResetThreads	endp


; ===============================================================
; = NBTHREADS                          				=
; =								=
; = Code for the forth word:    NbThreads              		=
; =								=
; = Stack:  ( --- n )                     			=
; ===============================================================

		align	4
NBTHREADS	proc

		; Push the number of active threads
		mov	eax, [ActiveThreadsNb]
		IPUSH	eax

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_NBTHREADS:
NBTHREADS	endp


; ===============================================================
; = ActiveThread                       				=
; =								=
; = Code for the forth word:    ActiveThread           		=
; =								=
; = Stack:  ( --- n )                     			=
; ===============================================================

		align	4
ACTIVETHREAD	proc

		; Push the current active thread
		mov	eax, [CurrActThread]
		sub	esi, @WordSize
		mov	[esi], eax

		; Verify stack frame overflow, then return to caller
		OVSTACK

		; Return to caller
		ret

E_ACTIVETHREAD:
ACTIVETHREAD	endp


; ===============================================================
; = NEWTHREAD                          				=
; =								=
; = Code for the forth word:    NewThread              		=
; =								=
; = Stack:  ( --- n )                     			=
; =								=
; = Input stream: NewThread name				=
; ===============================================================

		align	4
NEWTHREAD	proc

		; ----------------------------------------
		; Verify if the system is in compile state
		; ----------------------------------------

		call	QCOMPILE
		
		
		; -------------------------------------
		; Isolate next word in the input stream
		; -------------------------------------

		; Found next word in input stream buffer
		mov	edx, [IBUFFPTR2]
		inc	edx

		; Strip space chars: goto on first not blank char
@@LL2:		mov	al, [edx]
		cmpz	al, 0
		jne	@@LL6

		; Can't create a null word
		mov	ah, ERR_NULLWORD
		clr	edx					; No additionnal msg
		clr	ecx					; No line #
		call	doError

@@LL6:		cmp	al, 09h
		je	@@LL7
		cmp	al, ' '
		jne	@@LL3
@@LL7:		inc	edx
		jmp	@@LL2

@@LL3:
		; First, save current ptr: the start of the word
		mov	ecx, edx

		; Goto now on the first blank, or end of buffer
@@LL4:		mov	al, [ecx]
		cmpz	al, 0
		jz	@@LL5
		cmp	al, 09h
		je	@@LL5
		cmp	al, ' '
		je	@@LL5
		inc	ecx
		jmp	@@LL4

		; Found the end of the word
@@LL5:		mov	byte ptr [ecx], 0
		mov	[IBUFFPTR2], ecx		; Current pointer into the buffer


		; -------------------------------------------
		; Is this word defined into the dictionnary ?
		; -------------------------------------------
		
		sub	esi, @WordSize
		mov	[esi], edx
		call	PFind
		mov	eax, [esi]			; Get flag (word found ?)
		add	esi, @WordSize
		cmpz	eax, 0
		jne	@@WordDef
		
		; Error: word not defined
		mov	ah, ERR_WORDNDEF		; Word not defined
		clr	ecx				; No line #
		call	doError
		
@@WordDef:


ifdef NEVER
; !!!	
		; ---------------------
		; Update word cross ref
		; ---------------------

		; The word found NIA address (relative to NIA_START) is on top stack
		mov	[esi], ebx
		
		; Point to the last word defined in the current vocabulary: word currently being defined
		mov	eax, [VOC_CURRENT]
		add	eax, [VOC_START]
		mov	eax, [eax.voc_lastword]
		sub	esi, 4
		mov	[esi], eax
		
		; Update word cross ref
		call	CrossRefAdd
else
		add	esi, 4
endif


		; ---------------------------------
		; Compile the call to 'doNewThread'
		; ---------------------------------
		
		; Put 'SUMMON doNewThread' code operation
		compCALL_rel32	doNewThread

		; Put the 32-bits value into the Code Dictionnary
		add	ebx, [NIA_START]
		mov 	eax, [ebx.word_realaddr]
		sub	eax, [CDA_START]
		call	CodeComma               		; Code,
		

		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_NEWTHREAD:
NEWTHREAD	endp


; ===============================================================
; = doNewThread                        				=
; =								=
; = Code for the forth word:    doNewThread              	=
; =								=
; = Stack:  ( n --- )                     			=
; =								=
; = Input stream: NewThread name				=
; =								=
; = <n>	is the nb of 32-bits data to copy from current stack	=
; = 	to own thread parameters stack.				=
; ===============================================================

		align	4
doNewThread	proc


		; ------------------
		; Verify stack frame
		; ------------------
		
		VSTACK	1


		; ------------------
		; Too many threads ? 
		; ------------------
		
		cmp	[ActiveThreadsNb], MAXNB_THREADS
		jne	@@ThreadsNbOk

		; Error: too many threads
		mov	ebx, -1						; Push -1 on stack
		add	dword ptr [esp], @WordSize			; Skip next cell
		jmp	@@Done
		
@@ThreadsNbOk:


		; ------------------------------
		; Get the new thread code number
		; ------------------------------

		; Push the value (thread code address) on the parameters stack
		mov	eax, [esp]
		mov	eax, [eax]
		add	eax, [CDA_START]

		; Move the IP after the thread code address
		add	dword ptr [esp], @WordSize			; Skip next cell


		; ---------------------
		; Create the new thread
		; ---------------------
		
		; Update the number of active threads
		inc	[ActiveThreadsNb]
		
		; Found a free thread (inactive)
		lea	ecx, ThreadsActive
		clr	ebx
@@FindFreeThrd:	cmp	byte ptr [ecx + ebx], FALSE
		je	@@FindFreeThr2
		inc	ebx
		cmp	bx, MAXNB_THREADS
		jne	@@FindFreeThrd
@@FindFreeThr2:	mov	[ThreadsIP + ebx * 4], eax			; new thread IP

		; If 'EndThread' is not given, execute it by RET the calling thread word
		mov	eax, [SStackStart + ebx * 4]
		sub	eax, @WordSize
		lea	ecx, ENDTHREAD
		add	ecx, [OfsDictReloc]
		mov	dword ptr [eax], ecx
		mov	[ThreadsESP + ebx * 4], eax

		; This new thread is now actived
		mov	[ThreadsActive + ebx], TRUE


		; ---------------------------------------------------------------
		; Copy <n> 32-bits nb from current stack to this new thread stack
		; ---------------------------------------------------------------

		; Save used regs
		push	esi edi
		
		; Nb of bytes to move
		mov	ecx, [esi]
		inc	ecx
		
		; Source address
		mov	edi, [esi]
		lea	esi, [esi + edi * @WordSize + @WordSize]		; ESI = ESI + n * 4
		
		; Destination address
		mov	edi, [PStackStart + ebx * 4]
		
		; No copy ?
		cmp	ecx, 1
		je	@@NoCopy
		cmp	ecx, MAXNB_THREADS
		ja	@@NoCopy
		
		; Move data
		std                     				; Dec ESI and EDI
		rep	movsd  		            	    		; Move the block
		cld
@@NoCopy:

		; Restore used regs
		pop	edi esi


		; ----------------------------------
		; Start value of ESI for this thread
		; ----------------------------------

		mov	edx, [PARAM0]
		DISCARD	1
		mov	ecx, [PStackStart + ebx * 4]
		shl	edx, 2
		sub	ecx, edx
		mov	[ThreadsFSP + ebx * 4], ecx


		; -----------------------
		; Thread 1st passage flag
		; -----------------------

		mov	[Threads1stPass + ebx], TRUE			; 1st pass is true


		; --------------------------------------------
		; Push on stack the thread nb which is created 
		; --------------------------------------------
		
@@Done:		sub	esi, @WordSize
		mov	PARAM0, ebx

		
		; ----------------
		; Return to caller
		; ----------------
		
		ret

E_doNewThread:
doNewThread	endp


; ===============================================================
; = ENDTHREAD                          				=
; =								=
; = Code for the forth word:    EndThread              		=
; =								=
; = Stack:  ( --- )                     			=
; ===============================================================

		align	4
ENDTHREAD	proc

		; Do nothing if this is the thread #0
		mov	eax, [CurrActThread]
		cmpz	eax, 0
		je	@@Exit

		; This thread is disactived
		mov	[ThreadsActive + eax], FALSE

		; Update the number of nested loops for this thread
		mov	[LoopStackNb + eax * 4], 0

		; Update the number of nested case for this thread
		mov	[CaseStackNb + eax * 4], 0

		; Update the number of active threads
		dec	[ActiveThreadsNb]
		
		; Discard all local variables for this thread
		call	DiscardThreadLVar
		
		; Next active thread
@@Inactive:	inc	eax
		cmp	ax, MAXNB_THREADS
		jne	@@NotZero
		clr	eax
@@NotZero:	cmp	[ThreadsActive + eax], FALSE
		je	@@Inactive
		mov	[CurrActThread], eax

		; Update pointer on own parameters stack
		mov	esi, [ThreadsFSP + eax * 4]
		
		; Update pointer on own system stack
		mov	esp, [ThreadsESP + eax * 4]

		; Jump to next active thread saved IP
		jmp	[ThreadsIP + eax * 4]

		; Return to caller (if nothing to do)
@@Exit:		ret
		
E_ENDTHREAD:
ENDTHREAD	endp


; ===============================================================
; = KILLTHREAD                          			=
; =								=
; = Code for the forth word:    KillThread              	=
; =								=
; = Stack:  ( n --- )                     			=
; ===============================================================

		align	4
KILLTHREAD	proc

		; Verify stack frame
		VSTACK	1

		; Do nothing if this is the thread #0
		mov	eax, [esi]				; Get the nb of the thread to kill
		add	esi, @WordSize
		cmpz	eax, 0
		jle	@@Exit
		cmp	eax, MAXNB_THREADS
		jge	@@Exit

		; Is this thread active ?
		cmp	[ThreadsActive + eax], FALSE
		je	@@Exit
		
		; Discard all local variables for this thread
		call	DiscardThreadLVar
		
		; Is-it the current thread ?
		cmp	eax, [CurrActThread]
		je	ENDTHREAD

		; This thread is disactived
		mov	[ThreadsActive + eax], FALSE

		; Update the number of nested loops for this thread
		mov	[LoopStackNb + eax * 4], 0

		; Update the number of nested case for this thread
		mov	[CaseStackNb + eax * 4], 0

		; Update the number of active threads
		dec	[ActiveThreadsNb]
		
		; Return to caller (if nothing to do)
@@Exit:		ret
		
E_KILLTHREAD:
KILLTHREAD	endp


; ===============================================================
; = YIELD                          				=
; =								=
; = Code for the forth word:    YIELD              		=
; =								=
; = Stack:  ( --- )                     			=
; =								=
; = REGISTERS OTHERS THAN EAX ARE PRESERVED.			=
; ===============================================================

		align	4
YIELD		proc

		; Save used registers
		mov	[@@ecx], ecx

		; Save current stack pointer
		mov	eax, [CurrActThread]
		lea	ecx, [esp + @WordSize]
		mov	[ThreadsESP + eax * 4], ecx

		; Save current parameters stack pointer
		mov	[ThreadsFSP + eax * 4], esi
		
		; Save current thread IP
		mov	ecx, [esp]
		mov	[ThreadsIP + eax * 4], ecx

		; Next active thread
@@Inactive:	inc	eax
		cmp	ax, MAXNB_THREADS
		jne	@@NotZero
		clr	eax
@@NotZero:	cmp	[ThreadsActive + eax], FALSE
		je	@@Inactive
		mov	[CurrActThread], eax

		; Update pointer on own parameters stack
@@Not1stPass:	mov	esi, [ThreadsFSP + eax * 4]
@@Pass:

		; Restore initial value for used registers
		mov	ecx, [@@ecx]

		; Update pointer on own system stack
		mov	esp, [ThreadsESP + eax * 4]

		; Jump to next active thread saved IP
		jmp	[ThreadsIP + eax * 4]
		
		align	4
@@ecx		dd	?
		
E_YIELD:
YIELD		endp


code32		ends
		end
